import java.util.*;
import java.io.*;

public class EdgeDP_Paired_All {
	
	public static String[][] reads_1;
	public static String[][] reads_2;
    public static ArrayList[] align_1;
	public static ArrayList[] align_2;
	public static ArrayList[] lengths_1;
	public static ArrayList[] lengths_2;
	public static int[][] len_1;
	public static int[][] len_2;
    public static ArrayList[] confidence_1;
	public static ArrayList[] confidence_2;
    public static double[] c1;
	public static double[] c2;
    public static Integer[][] pos_1;
	public static Integer[][] pos_2;
	public static int fasta_readlen;
	public static double beta = 1.0;
	
	public static SpliceGraph s1;
	public static SpliceGraph s2;
	
	public static double[] z_1;
	public static double[] z_2;
	
	public static Edge[] edge_list_1;
	public static Edge[] edge_list_2;
	public static Segment[] segments;
	public static Edge[][] edges;
	
    public static void main(String[] args) {
		if (args.length != 6 && args.length != 9) {
			System.out.println("usage: EdgeDP_Paired <psg1> <reads1.1> <reads1.2>" +
							   " <reads2.1> <reads2.2> <fasta_readlen>"+
							   " [<mean> <sd> <beta>]");
			System.exit(0);
		}
		
		String p1 = args[0];
		String r1_1 = args[1];
		String r1_2 = args[2];
		String r2_1 = args[3];
		String r2_2 = args[4];
		
		fasta_readlen = Integer.parseInt(args[5]);
		int verbose = 0;
		
		s1 = new SpliceGraph(p1);
		s2 = new SpliceGraph(p1);
		segments = s1.getSegments();
		
		process_reads(r1_1,r1_2,1);
		process_reads(r2_1,r2_2,2);
		
		try {
			int mean = Integer.parseInt(args[6]);
			s1.setMean(mean);
			s2.setMean(mean);
			double sd = Double.parseDouble(args[7]);
			s1.setStdev(sd);
			s2.setStdev(sd);
			beta = Double.parseDouble(args[8]);
			s1.setBeta(beta);
			s2.setBeta(beta);
		}
		catch (ArrayIndexOutOfBoundsException a) {
			// do nothing, this is fine
		}
		catch (NumberFormatException nfe) {
			// do nothing, we'll just run with the defaults
		}
		
		s1.resetEdgeWeights();
		len_1 = Utility.IntArray(lengths_1);
		s1.learnEM(reads_1,align_1,confidence_1,verbose,len_1);
		//double[] weights_1 = s1.getEdgeWeights();
		double ll_1 = s1.getLogLikelihood(reads_1,align_1,confidence_1,len_1);
		
		s2.resetEdgeWeights();
		len_2 = Utility.IntArray(lengths_2);
		s2.learnEM(reads_2,align_2,confidence_2,verbose,len_2);
		//double[] weights_2 = s2.getEdgeWeights();
		double ll_2 = s2.getLogLikelihood(reads_2,align_2,confidence_2,len_2);
		
		double alternative = ll_1+ll_2;
		
		//System.out.println("Alternative Log Likelihood: "+alternative);
		
		System.out.printf(" Vertex | Alternative | Null | df\n");
		System.out.printf("------------------------------------\n");
		for (int i=0; i<segments.length; i++) {
			if (s1.get_outdegree(segments[i]) > 1) {
				s1.resetEdgeWeights();
				s2.resetEdgeWeights();
				double nullH = learnEM(i);
				//System.out.println("Null log likelihood: "+nullH);
				System.out.printf("%4d    | %8.3f | %3.3f | %2d\n",i,alternative,nullH,s1.get_outdegree(segments[i])-1);
			}
		}
    }
	
	private static double learnEM(int vertex) {
		// convergence conditions
		double min_edge_up = .00001, min_inc_ll = .001;
		boolean done = false;
		edge_list_1 = s1.getEdgeList();
		edge_list_2 = s2.getEdgeList();
		edges = s1.getEdges();
		
		// things to check
		double ll_1 = s1.eStep(align_1,len_1,confidence_1);
		double ll_2 = s2.eStep(align_2,len_2,confidence_2);
		double ll = ll_1+ll_2;
		int iter = 0;
		double wgt_update = 0;
		//s1.emOutput(ll_1, iter);
		//s2.emOutput(ll_2, iter);
		
		z_1 = s1.getZ();
		z_2 = s2.getZ();
		
		float num_reads_1 = 0;
		float num_reads_2 = 0;
		for (int i=0; i<edges[0].length; i++) {
			if (edges[0][i] != null) {
				int index = s1.getEdgeIndex(edges[0][i]);
				num_reads_1 += z_1[index];
				num_reads_2 += z_2[index];
			}
		}
		
		while (!done) {
			//double prev_ll_1 = ll_1;
			//double prev_ll_2 = ll_2;
			double prev_ll = ll_1 + ll_2;
			double[] prev_probs_1 = new double[edge_list_1.length];
			double[] prev_probs_2 = new double[edge_list_2.length];
			for (int i=0; i<edge_list_1.length; i++) {
				prev_probs_1[i] = edge_list_1[i].weight;
			}
			for (int i=0; i<edge_list_2.length; i++) {
				prev_probs_2[i] = edge_list_2[i].weight;
			}
			
			mStep(Math.round(num_reads_1),Math.round(num_reads_2),vertex);
			ll_1 = s1.eStep(align_1,len_1,confidence_1);
			ll_2 = s2.eStep(align_2,len_2,confidence_2);
			wgt_update = getWeightUpdate(prev_probs_1,prev_probs_2);
			ll = ll_1 + ll_2;
			//s1.emOutput(ll_1, iter);
			//s2.emOutput(ll_2, iter);
			
			if ((ll-prev_ll) > min_inc_ll /*&& (ll_2-prev_ll_2) > min_inc_ll*/
				&& wgt_update > min_edge_up) {
				iter++;
			}
			else {
				iter++;
				if (ll-prev_ll < 0 && Math.abs(ll-prev_ll) > min_inc_ll) {
					System.err.println("ERROR: decreasing likelihood");
					System.err.println("  Number of reads: "+reads_1.length+","+reads_2.length);
					System.err.println("  Iterations: "+iter);
				}
				done = true;
			}
		}
		return ll;
	}
	
	// Maximization step with bonus priors
    public static void mStep(int num_reads_1, int num_reads_2, int vertex) {
		double update = 0, min_update = .001;
		double l_update = 0;
		segments = s1.getSegments();
		double[] lambda_1 = s1.getLambda();
		double[] lambda_2 = s2.getLambda();
		double[] old_wgts_1 = new double[edge_list_1.length];
		double[] old_lamb_1 = new double[lambda_1.length];
		double[] old_wgts_2 = new double[edge_list_2.length];
		double[] old_lamb_2 = new double[lambda_2.length];
		boolean[] execute_1 = new boolean[segments.length];
		boolean[] execute_2 = new boolean[segments.length];
		do {
			double[] avg_pr_1 = s1.d_avg_pref();
			double[] avg_su_1 = s1.d_avg_suff();
			double[] avg_pr_2 = s2.d_avg_pref();
			double[] avg_su_2 = s2.d_avg_suff();
			z_1 = s1.getZ();
			z_2 = s2.getZ();
			double[][] f_1 = s1.getF();
			double[][] f_2 = s2.getF();
			for (int i=0; i<execute_1.length; i++) {
				execute_1[i] = true;
			}
			for (int i=0; i<execute_2.length; i++) {
				execute_2[i] = true;
			}
			double[] denoms_1 = new double[edge_list_1.length];
			double[] numers_1 = new double[edge_list_1.length];
			double[] denoms_2 = new double[edge_list_2.length];
			double[] numers_2 = new double[edge_list_2.length];
			
			double[] alpha_bar_1 = new double[edge_list_1.length];
			double[] alpha_bar_2 = new double[edge_list_2.length];
			
			double[] new_wgts_1 = new double[edge_list_1.length];
			double[] new_wgts_2 = new double[edge_list_2.length];
			double[] new_lamb_1 = new double[segments.length-1];
			double[] new_lamb_2 = new double[segments.length-1];
			
			// find \bar{alpha_ij} = c_ij + b_ij/(N/D(a)f(i)(d_p(i)+d_q(j))+lambda_i)
			for (int i=0; i<edge_list_1.length; i++) {
				Edge e = edge_list_1[i];
				int source = s1.getSegIndex(e.source);
				int target = s1.getSegIndex(e.target);
				if (source == vertex) {
					numers_1[i] = z_1[i] + z_2[i] + beta;
					numers_2[i] = z_1[i] + z_2[i] + beta;
					denoms_1[i] = (num_reads_1/avg_su_1[0])*f_1[0][source]*(avg_pr_1[source]+avg_su_1[target]);
					denoms_1[i] += (num_reads_2/avg_su_2[0])*f_2[0][source]*(avg_pr_2[source]+avg_su_2[target]);
					denoms_1[i] += lambda_1[source];
					denoms_2[i] = (num_reads_1/avg_su_1[0])*f_1[0][source]*(avg_pr_1[source]+avg_su_1[target]);
					denoms_2[i] += (num_reads_2/avg_su_2[0])*f_2[0][source]*(avg_pr_2[source]+avg_su_2[target]);
					denoms_2[i] += lambda_2[source];
				}
				else {
					numers_1[i] = z_1[i] + beta;
					numers_2[i] = z_2[i] + beta;
					denoms_1[i] = (num_reads_1/avg_su_1[0])*f_1[0][source]*(avg_pr_1[source]+avg_su_1[target])+lambda_1[source];
					denoms_2[i] = (num_reads_2/avg_su_2[0])*f_2[0][source]*(avg_pr_2[source]+avg_su_2[target])+lambda_2[source];
				}
				alpha_bar_1[i] = numers_1[i]/denoms_1[i];
				alpha_bar_2[i] = numers_2[i]/denoms_2[i];
				if (denoms_1[i] <= 0 || Double.isNaN(denoms_1[i]) /*|| numers_1[i] == 0*/) {
					execute_1[source] = false;
				}
				if (denoms_2[i] <= 0 || Double.isNaN(denoms_2[i]) /*|| numers_2[i] == 0*/) {
					execute_2[source] = false;
				}
			}
			//*
			for (int i=0; i<edges.length; i++) {
				double total_numerator_1 = 0.0;
				double total_numerator_2 = 0.0;
				for (int j=0; j<edges[i].length; j++) {
					if (edges[i][j] != null) {
						total_numerator_1 += numers_1[s1.getEdgeIndex(edges[i][j])];
						total_numerator_2 += numers_2[s1.getEdgeIndex(edges[i][j])];
					}
				}
				if (total_numerator_1 <= 0.0) {
					execute_1[i] = false;
 				}
				if (total_numerator_2 <= 0.0) {
					execute_2[i] = false;
 				}
				//System.err.println("seg "+i+": "+total_numerator);
			}
			// */
			// find alpha_ij = \bar{alpha_ij}/sum_j alpha_ijt
			for (int i=0; i<edge_list_1.length; i++) {
				Edge e = edge_list_1[i];
				int source = s1.getSegIndex(e.source);
				double denominator_1 = 0.0;
				double denominator_2 = 0.0;
				for (int j=0; j<edges[source].length; j++) {
					if (edges[source][j] != null) {
						denominator_1 += alpha_bar_1[s1.getEdgeIndex(edges[source][j])];
						denominator_2 += alpha_bar_2[s1.getEdgeIndex(edges[source][j])];
					}
				}
				old_wgts_1[i] = e.weight;
				old_wgts_2[i] = edge_list_2[i].weight;
				if (execute_1[source]) {
					new_wgts_1[i] = alpha_bar_1[i]/denominator_1;
				}
				else {
					// updated logic: do not update
				}
				if (execute_2[source]) {
					new_wgts_2[i] = alpha_bar_2[i]/denominator_2;
				}
				else {
					// updated logic: do not update
				}
			}
			s1.setEdgeWeights(new_wgts_1);
			s2.setEdgeWeights(new_wgts_2);
			update = getWeightUpdate(old_wgts_1,old_wgts_2);
			
			// update lambda_i = c_i + b_i - N/D(a)f(i)(d_p(i)+d_q(j)-ell(i))
			s1.calculate_f();
			s2.calculate_f();
			f_1 = s1.getF();
			f_2 = s2.getF();
			for (int i=0; i<lambda_1.length; i++) {
				double c_i1 = 0.0;
				double b_i1 = 0.0;
				double c_i2 = 0.0;
				double b_i2 = 0.0;
				for (int j=0; j<edges[i].length; j++) {
					if (edges[i][j] != null) {
						int index = s1.getEdgeIndex(edges[i][j]);
						c_i1 += z_1[index];
						b_i1 += beta;
						c_i2 += z_2[index];
						b_i2 += beta;
					}
				}
				old_lamb_1[i] = lambda_1[i];
				old_lamb_2[i] = lambda_2[i];
				if (i == vertex) {
					
				}
				else {
					new_lamb_1[i] = c_i1 + b_i1 - (num_reads_1/avg_su_1[0])*f_1[0][i]*(avg_pr_1[i]+avg_su_1[i]-segments[i].length());
					new_lamb_2[i] = c_i2 + b_i2 - (num_reads_2/avg_su_2[0])*f_2[0][i]*(avg_pr_2[i]+avg_su_2[i]-segments[i].length());
				}
			}
			s1.setLambda(new_lamb_1);
			s2.setLambda(new_lamb_2);
			l_update = getLambdaUpdate(old_lamb_1,old_lamb_2);
		} while (update > min_update || l_update > min_update);
    }
	
	public static double getWeightUpdate(double[] oldWgts_1, double[] oldWgts_2) {
		edge_list_1 = s1.getEdgeList();
		edge_list_2 = s2.getEdgeList();
		double update = 0;
		for (int i=0; i<oldWgts_1.length; i++) {
			double diff = Math.abs(oldWgts_1[i]-edge_list_1[i].weight);
			if (diff > update) {
				update = diff;
			}
		}
		for (int i=0; i<oldWgts_2.length; i++) {
			double diff = Math.abs(oldWgts_2[i]-edge_list_2[i].weight);
			if (diff > update) {
				update = diff;
			}
		}
		return update;
	}
	
	private static double getLambdaUpdate(double[] oldWgts_1, double[] oldWgts_2) {
		double[] lambda_1 = s1.getLambda();
		double[] lambda_2 = s2.getLambda();
		double update = 0;
		for (int i=0; i<oldWgts_1.length; i++) {
			double diff = Math.abs(oldWgts_1[i]-lambda_1[i]);
			if (diff > update) {
				update = diff;
			}
		}
		for (int i=0; i<oldWgts_2.length; i++) {
			double diff = Math.abs(oldWgts_2[i]-lambda_2[i]);
			if (diff > update) {
				update = diff;
			}
		}
		return update;
	}
	
	private static double bic(String[] reads, double ll) {
		// BIC = ln p(D|\theta_ML) - 0.5 M ln N
		// M = (\sum_v outdegree(v)) - |V| + 1
		double M = edge_list_1.length;
		double N = reads.length;
		M -= segments.length;
		M++;
		//System.out.println("BIC="+ll+"-0.5*"+M+"*ln("+N+")");
		return (ll-0.5*M*Math.log(N));
	}
	
	// reads information from a bowtie alignment file
    private static void process_reads(String filename1, String filename2, int which) {
		// initialize the temporary variables
		ArrayList<Integer> start1 = new ArrayList<Integer>();
		ArrayList<Integer> start2 = new ArrayList<Integer>();
		ArrayList<String> read1 = new ArrayList<String>();
		ArrayList<String> read2 = new ArrayList<String>();
		ArrayList<String> alig1 = new ArrayList<String>();
		ArrayList<String> alig2 = new ArrayList<String>();
		ArrayList<String> which1 = new ArrayList<String>();
		ArrayList<String> which2 = new ArrayList<String>();
		
		ArrayList<String> mismatch = new ArrayList<String>();
		ArrayList<String> quality = new ArrayList<String>();
		
		// initialize the global variables
		if (which == 1) {
			reads_1 = null;
			align_1 = null;
			confidence_1 = null;
		}
		else {
			reads_2 = null;
			align_2 = null;
			confidence_2 = null;
		}
		c1 = null;
		c2 = null;
		
		try {
			Scanner r = new Scanner(new File(filename1));
			while (r.hasNextLine()) {
				String[] components = r.nextLine().split("\t");
				which1.add(components[0]);
				alig1.add(components[2]);
				start1.add(Integer.decode(components[3]));
				read1.add(components[4]);
				quality.add(components[5]);
				if (components.length == 8) {
					mismatch.add(components[7]);
				}
				else {
					mismatch.add("x");
				}
			}
		}
		catch (IOException ioe) {
			System.out.println("Alignment file "+filename1+" contains an error.");
			System.exit(0);
		}
		c1 = compute_quality(quality,mismatch);
		
		mismatch = new ArrayList<String>();
		quality = new ArrayList<String>();
		
		try {
			Scanner r = new Scanner(new File(filename2));
			while (r.hasNextLine()) {
				String[] components = r.nextLine().split("\t");
				which2.add(components[0]);
				alig2.add(components[2]);
				start2.add(Integer.decode(components[3]));
				read2.add(components[4]);
				quality.add(components[5]);
				if (components.length == 8) {
					mismatch.add(components[7]);
				}
				else {
					mismatch.add("x");
				}
			}
		}
		catch (IOException ioe) {
			System.out.println("Alignment file "+filename2+" contains an error.");
			System.exit(0);
		}
		c2 = compute_quality(quality,mismatch);
		//System.err.println("\tQuality computed.");
		
		if (which == 1) {
			reads_1 = getReads(which1,which2,read1,read2);
		}
		else {
			reads_2 = getReads(which1,which2,read1,read2);
		}
		//System.err.println("\tReads got.");
		
		if (which == 1 && reads_1.length > 0) {
			process_align(which1,which2,alig1,alig2,start1,start2,which);
			//System.err.println("\tAlignments processed.");
		}
		else if (which == 2 && reads_2.length > 0) {
			process_align(which1,which2,alig1,alig2,start1,start2,which);
		}
    }
    
	// attempting to salvage this method
	private static void process_align(ArrayList<String> id1, ArrayList<String> id2, 
									  ArrayList<String> al1, ArrayList<String> al2,
									  ArrayList<Integer> st1, ArrayList<Integer> st2,
									  int which) {
		if (which == 1) {
			align_1 = new ArrayList[reads_1.length];
			confidence_1 = new ArrayList[reads_1.length];
			lengths_1 = new ArrayList[reads_1.length];
		}
		else {
			align_2 = new ArrayList[reads_2.length];
			confidence_2 = new ArrayList[reads_2.length];
			lengths_2 = new ArrayList[reads_2.length];
		}
		
		int index = 0;
		int old_ind = -1;
		
		// one for each of the paired reads
		ArrayList<Integer[]> matches1 = new ArrayList<Integer[]>();
		ArrayList<Double> quality1 = new ArrayList<Double>();
		ArrayList<Integer> pos1 = new ArrayList<Integer>();
		ArrayList<Integer[]> matches2 = new ArrayList<Integer[]>();
		ArrayList<Double> quality2 = new ArrayList<Double>();
		ArrayList<Integer> pos2 = new ArrayList<Integer>();
		
		HashMap<String,Integer> used_ids = new HashMap<String,Integer>();
		
		for (int i=0; i<id1.size(); i++) {
			// initialize everything
			String id = id1.get(i);
			if (used_ids.containsKey(id)) {
				old_ind = used_ids.get(id).intValue();
			}
			matches1 = new ArrayList<Integer[]>();
			quality1 = new ArrayList<Double>();
			pos1 = new ArrayList<Integer>();
			
			matches2 = new ArrayList<Integer[]>();
			quality2 = new ArrayList<Double>();
			pos2 = new ArrayList<Integer>();
			
			// get first read alignment and trim it
			String[] match = al1.get(i).substring(1).split(":");
			Integer[] r = new Integer[match.length];
			for (int j=0; j<match.length; j++) {
				r[j] = Integer.decode(match[j]);
			}
			Integer[] t = null;
			t = adjust_alignment(r,st1.get(i).intValue(),reads_1[0][0].length());
			int k=0;
			int new_st1 = st1.get(i);
			if (r[k].intValue() != t[0].intValue()) {
				new_st1 -= Math.min(s1.getSegment(r[k].intValue()).length(),fasta_readlen-1);
				k++;
			}
			else {
				if (r.length == 1) {
					// we're fine, only aligned to one exon anyway
				}
				else {
					// it's an intron-bridging alignment, add the rest of the segment in
					new_st1 += s1.getSegment(r[k].intValue()).length() - (fasta_readlen - 1);
				}
			}
			while (r[k].intValue() != t[0].intValue() && new_st1 > 0) {
				new_st1 -= s1.getSegment(r[k].intValue()).length();
				k++;
			}
			
			if (!contains(matches1,r)) {
				matches1.add(r);
				quality1.add(c1[i]);
				pos1.add(new_st1);
			}
			
			id = id.substring(0,id.length()-1);
			id += "2";
			for (int j=0; j<id2.size(); j++) {
				if (id2.get(j).equals(id)) {
					// we have a second half of this read
					// get its alignment and trim it
					String[] m = al2.get(j).substring(1).split(":");
					r = new Integer[m.length];
					for (k=0; k<m.length; k++) {
						r[k] = Integer.decode(m[k]);
					}
					
					t = adjust_alignment(r,st2.get(j).intValue(),reads_1[0][0].length());
					k=0;
					int new_st2 = st2.get(j);
					if (r[k].intValue() != t[0].intValue()) {
						new_st2 -= Math.min(s1.getSegment(r[k].intValue()).length(),fasta_readlen-1);
						k++;
					}
					else {
						if (r.length == 1) {
							// we're fine, only aligned to one exon anyway
						}
						else {
							// it's an intron-bridging alignment, add the rest of the segment in
							new_st2 += s1.getSegment(r[k].intValue()).length() - (fasta_readlen - 1);
						}
					}
					while (r[k].intValue() != t[0].intValue()) {
						new_st2 -= s1.getSegment(r[k].intValue()).length();
						k++;
					}
					r = t;
					
					if (!contains(matches2,r)) {
						matches2.add(r);
						quality2.add(c2[j]);
						pos2.add(new_st2);
					}
				}
			}
			ArrayList[] paths = s1.find_paths(matches1,matches2,quality1,quality2,pos1,pos2,
											  reads_1[0][0].length());
			
			if (old_ind == -1 && paths != null && which == 1) {
				//System.out.print(index + " ");
				used_ids.put(id1.get(i), index);
				align_1[index] = paths[0];
				confidence_1[index] = paths[1];
				lengths_1[index] = paths[2];
				index++;
			}
			else if (old_ind == -1 && paths != null && which == 2) {
				//System.out.print(index + " ");
				used_ids.put(id1.get(i), index);
				align_2[index] = paths[0];
				confidence_2[index] = paths[1];
				lengths_2[index] = paths[2];
				index++;
			}
			else if (paths != null && which == 1) {
				for (int j=0; j<paths[0].size(); j++) {
					align_1[old_ind].add(paths[0].get(j));
				}
				for (int j=0; j<paths[1].size(); j++) {
					confidence_1[old_ind].add(paths[1].get(j));
				}
				for (int j=0; j<paths[2].size(); j++) {
					lengths_1[old_ind].add(paths[2].get(j));
				}
				//align[old_ind] = matches;
				//confidence[old_ind] = quality;
				old_ind = -1;
			}
			else if (paths != null && which == 2) {
				for (int j=0; j<paths[0].size(); j++) {
					align_2[old_ind].add(paths[0].get(j));
				}
				for (int j=0; j<paths[1].size(); j++) {
					confidence_2[old_ind].add(paths[1].get(j));
				}
				for (int j=0; j<paths[2].size(); j++) {
					lengths_2[old_ind].add(paths[2].get(j));
				}
				//align[old_ind] = matches;
				//confidence[old_ind] = quality;
				old_ind = -1;
			}
		}
		//System.out.println("#");
    }
	
	// check whether this alignment needs to be trimmed
	// and do so if necessary
	private static Integer[] adjust_alignment(Integer[] r, int pos, int read_len) {
		if (r.length == 1) {
			// st.get(i) is an index into that segment
			// do nothing, this is as simple as it gets
		}
		else {
			// find first segment used
			int j=0;
			Segment temp_1 = s1.getSegment(r[0]);
			if (pos > Math.min(temp_1.length(),fasta_readlen)) {
				pos -= Math.min(temp_1.length(),fasta_readlen-1);
				j++;
				temp_1 = s1.getSegment(r[j]);
				while (pos > temp_1.length()) {
					if (temp_1.length() > 0) {
						pos -= Math.min(temp_1.length(),fasta_readlen-1);
						j++;
						temp_1 = s1.getSegment(r[j]);
					}
					else {
						pos = 0;
					}
				}
			}
			
			// pos is now an index into the jth segment in r
			// (subsegment if j=0)
			int total = 0;
			if (j == 0) {
				total += Math.min(s1.getSegment(r[0]).length(),fasta_readlen-1)-pos;
			}
			else {
				total += s1.getSegment(r[j]).length()-pos;
			}
			int k=j;
			if (total >= read_len || s1.getSegment(r[j]).length() == 0) {
				// then j is the only relevant segment
			}
			else {
				// let's find the end!
				while (total < read_len) {
					k++;
					temp_1 = s1.getSegment(r[k]);
					if (temp_1.length() > 0) {
						total += temp_1.length();
					}
					else {
						// we've arrived at the poly-A tail!
						total = read_len;
					}
				}
			}
			
			// trim r between j and k
			Integer[] temp = new Integer[k-j+1];
			int ind = 0;
			for (int x=j; x<=k; x++) {
				temp[ind] = r[x];
				ind++;
			}
			r = temp;
		}
		return r;
	}
    
	// retrieves the reads from the read arraylist
	private static String[][] getReads(ArrayList<String> r1, ArrayList<String> r2,
									   ArrayList<String> s1, ArrayList<String> s2) {
		int num1 = count_unique(r1);
		int num2 = count_unique(r2);
		String[][] result = new String[Math.min(num1,num2)][2];
		String id = null;
		int index = 0;
		ArrayList<String> used_ids = new ArrayList<String>();
		
		for (int i=0; i<r1.size(); i++) {
			if (!contains(used_ids, r1.get(i))) {
				id = r1.get(i);
				used_ids.add(id);
				//result[index][0] = s1.get(i);
				id = id.substring(0,id.length()-1);
				id += "2";
				for (int j=0; j<r2.size(); j++) {
					if (r2.get(j).equals(id)) {
						result[index][0] = s1.get(i);
						result[index][1] = s2.get(j);
						//index++;
					}
				}
				//*
				 if (result[index][1] != null) {
				 index++;
				 }
				 // */
			}
		}
		return trim(result);
    }
	
	private static String[][] trim(String[][] to_trim) {
		int i=0;
		for (i=0; i<to_trim.length; i++) {
			if (to_trim[i][1] == null) {
				break;
			}
		}
		String[][] result = new String[i][];
		for (i=0; i<result.length; i++) {
			result[i] = to_trim[i];
		}
		return result;
    }
	
	// Computes confidence based on match/mismatch; quality of read
    private static double[] compute_quality(ArrayList<String> q, ArrayList<String> m) {
		double[] result = new double[q.size()];
		for (int i=0; i<q.size(); i++) {
			result[i] = 1; // initializing for multiplication
			// check for mismatches
			int[] positions = {-1};
			if (!m.get(i).equals("x")) {
				// process mismatches
				String[] mismatch = m.get(i).split(",");
				positions = new int[mismatch.length];
				for (int j=0; j<mismatch.length; j++) {
					// get position of mismatch - part before : should be position
					String[] temp = mismatch[j].split(":");
					positions[j] = Integer.parseInt(temp[0]);
				}
			}
			// process quality scores
			int index = 0; // for positions[], assume listed in increasing order
			for (int j=0; j<q.get(i).length(); j++) {
				int quality = (int) q.get(i).charAt(j);
				quality -= 33; // because we have a +33 offset
				if (index >= positions.length || j != positions[index]) {
					// not a mismatch
					result[i] *= 1-Math.pow(10,(-quality/10.0));
				}
				else {
					// mismatch
					result[i] *= (1/3.0)*Math.pow(10,(-quality/10.0));
					index++;
				}
			}
		}
		return result;
    }
	
	// retrieves the reads from the read arraylist
    private static String[] getReads(ArrayList<String> r, ArrayList<String> s) {
		int num = count_unique(r);
		String[] result = new String[num];
		String id = null;
		int index = 0;
		ArrayList<String> used_ids = new ArrayList<String>();
		for (int i=0; i<r.size(); i++) {
			if (!contains(used_ids, r.get(i))) {
				id = r.get(i);
				used_ids.add(id);
				result[index] = s.get(i);
				index++;
			}
		}
		return trim(result);
    }
	
	private static String[] trim(String[] to_trim) {
		int i=0;
		for (i=0; i<to_trim.length; i++) {
			if (to_trim[i] == null) {
				break;
			}
		}
		String[] result = new String[i];
		for (i=0; i<result.length; i++) {
			result[i] = to_trim[i];
		}
		return result;
    }
	
	private static void process_align(int l, ArrayList<String> id, 
									  ArrayList<String> al, double[] c,
									  int read_len, ArrayList<Integer> st,
									  int ind) {
		if (ind == 0) {
			align_1 = new ArrayList[l];
			confidence_1 = new ArrayList[l];
		}
		else {
			align_2 = new ArrayList[l];
			confidence_2 = new ArrayList[l];
		}
		
		int index = 0;
		int old_ind = -1;
		ArrayList<Integer[]> matches = new ArrayList<Integer[]>();
		ArrayList<Double> quality = new ArrayList<Double>();
		
		HashMap<String,Integer> used_ids = new HashMap<String,Integer>();
		
		for (int i=0; i<id.size(); i++) {
			if (!used_ids.containsKey(id.get(i))) {
				//used_ids.put(id.get(i), index);
				matches = new ArrayList<Integer[]>();
				quality = new ArrayList<Double>();
			}
			else {
				old_ind = used_ids.get(id.get(i)).intValue();
				if (ind == 0) {
					matches = (ArrayList<Integer[]>) align_1[old_ind];
					quality = (ArrayList<Double>) confidence_1[old_ind];
				}
				else {
					matches = (ArrayList<Integer[]>) align_2[old_ind];
					quality = (ArrayList<Double>) confidence_2[old_ind];
				}
			}
			String[] match = al.get(i).substring(1).split(":");
			Integer[] r = new Integer[match.length];
			for (int j=0; j<match.length; j++) {
				r[j] = Integer.decode(match[j]);
			}
			// check whether this alignment needs to be trimmed
			// and do so if necessary
			if (r.length == 1) {
				// st.get(i) is an index into that segment
				// do nothing, this is as simple as it gets
			}
			else {
				// find first segment used
				int j=0;
				int pos = st.get(i).intValue();
				Segment temp_1 = s1.getSegment(r[0]);
				if (pos > Math.min(temp_1.length(),fasta_readlen)) {
					pos -= Math.min(temp_1.length(),fasta_readlen-1);
					j++;
					temp_1 = s1.getSegment(r[j]);
					while (pos > temp_1.length()) {
						if (temp_1.length() > 0) {
							pos -= Math.min(temp_1.length(),fasta_readlen-1);
							j++;
							temp_1 = s1.getSegment(r[j]);
						}
						else {
							pos = 0;
						}
					}
				}
				
				// pos is now an index into the jth segment in r
				// (subsegment if j=0)
				int total = 0;
				if (j == 0) {
					total += Math.min(s1.getSegment(r[0]).length(),fasta_readlen-1)-pos;
				}
				else {
					total += s1.getSegment(r[j]).length()-pos;
				}
				int k=j;
				if (total >= read_len || s1.getSegment(r[j]).length() == 0) {
					// then j is the only relevant segment
				}
				else {
					// let's find the end!
					while (total < read_len) {
						k++;
						temp_1 = s1.getSegment(r[k]);
						if (temp_1.length() > 0) {
							total += temp_1.length();
						}
						else {
							// we've arrived at the poly-A tail!
							total = read_len;
						}
					}
				}
				
				// trim r between j and k
				Integer[] temp = new Integer[k-j+1];
				int derp = 0;
				for (int x=j; x<=k; x++) {
					temp[derp] = r[x];
					derp++;
				}
				r = temp;
				
			}
			if (!contains(matches,r)) {
				matches.add(r);
				quality.add(c[i]);
			}
			
			if (!used_ids.containsKey(id.get(i))) {
				used_ids.put(id.get(i), index);
				if (ind == 0) {
					align_1[index] = matches;
					confidence_1[index] = quality;
				}
				else {
					align_2[index] = matches;
					confidence_2[index] = quality;
				}
				index++;
			}
			else {
				if (ind == 0) {
					align_1[old_ind] = matches;
					confidence_1[old_ind] = quality;
				}
				else {
					align_2[old_ind] = matches;
					confidence_2[old_ind] = quality;
				}
			}
		}
    }
	
	// Assumption: any duplicates will be consecutive
    private static int count_unique(ArrayList<String> r) {
		if (r.size() == 0) { return 0; }
		int unique = 1;
		int index = 1;
		String current = r.get(0);
		while (index < r.size()) {
			if (!r.get(index).equals(current)) {
				current = r.get(index);
				unique++;
			}
			index++;
		}
		return unique;
    }
	
	private static boolean contains(ArrayList<String> a, String b) {
		boolean retval = false;
		for (int i=0; i<a.size(); i++) {
			if (a.get(i).equals(b)) {
				retval = true;
			}
		}
		return retval;
    }
    
    private static boolean contains(ArrayList<Integer[]> a, Integer[] b) {
		boolean retval = false;
		for (int i=0; i<a.size(); i++) {
			if (a.get(i).length == b.length) {
				boolean verify = true;
				for (int j=0; j<a.get(i).length; j++) {
					if (a.get(i)[j] != b[j]) {
						verify = false;
					}
				}
				if (verify) {
					retval = true;
					break;
				}
			}
		}
		return retval;
    }
	
}