/*
 * Here we make the assumption that the .psg and .em.out files are in a given
 * file structure. Cuz I say so.
 */

import java.util.*;
import java.io.*;

public class FactorCompare {
	
	public static SpliceGraph s_exon;
	public static SpliceGraph s_line;
	public static SpliceGraph s_full;
	public static SpliceGraph s_unfac;
    
    public static void main(String[] args) {
		if (args.length != 1) {
			System.out.println("usage: FactorCompare <gene_name>");
			System.exit(0);
		}
		
		String gene_name = args[0];
		s_exon = new SpliceGraph("TestPSG/factorization/exon/"+args[0]+".psg");
		s_line = new SpliceGraph("TestPSG/factorization/line/"+args[0]+".psg");
		s_full = new SpliceGraph("TestPSG/factorization/full_length/"+args[0]+".psg");
		s_unfac = new SpliceGraph("TestPSG/factorization/unfactorized/"+args[0]+".psg");
		
		get_values(s_line,"TestPSG/factorization/line/"+args[0]+".em.out");
		System.out.println("== LINE GRAPH ==");
		s_line.printEdgeList();
		System.out.println();
		get_values(s_exon,"TestPSG/factorization/exon/"+args[0]+".em.out");
		//s_exon.printEdgeList();
		//System.out.println();
		get_values(s_unfac,"TestPSG/factorization/unfactorized/"+args[0]+".em.out");
		//s_unfac.printEdgeList();
		//System.out.println();
		get_values(s_full,"TestPSG/factorization/full_length/"+args[0]+".em.out");
		//s_full.printEdgeList();
		//System.out.println();
		
		tabulate();
    }
	
	private static void tabulate() {
		// Step 1: get all of the line graph segment coordinates
		HashMap<Integer,Integer> start_pos = new HashMap<Integer,Integer>();
		HashMap<Integer,Integer> end_pos = new HashMap<Integer,Integer>();
		Segment[] segs = s_line.getSegments();
		for (int i=0; i<segs.length; i++) {
			if (i != 0) {
				start_pos.put(segs[i].start,i);
			}
			if (i != segs.length-1) {
				end_pos.put(segs[i].end,i);
			}
		}
		
		// Step 2: for each graph structure, map the edges to the line graph
		int[] exon_map = match_edges(s_exon,start_pos,end_pos);
		int[] full_map = match_edges(s_full,start_pos,end_pos);
		int[] unfac_map = match_edges(s_unfac,start_pos,end_pos);
		
		// Step 3: add the weights
		double[] exon_weight = sum_weights(s_exon, exon_map);
		double[] full_weight = sum_weights(s_full, full_map);
		double[] unfac_weight = sum_weights(s_unfac, unfac_map);
		
		System.out.println("== EXON GRAPH ==");
		for (int i=0; i<exon_weight.length; i++) {
			System.out.printf("%.2f ",exon_weight[i]);
		}
		System.out.println();
		System.out.println();
		System.out.println("== FULL GRAPH ==");
		for (int i=0; i<full_weight.length; i++) {
			System.out.printf("%.2f ",full_weight[i]);
		}
		System.out.println();
		System.out.println();
		System.out.println("== UNFACTORIZED GRAPH ==");
		for (int i=0; i<unfac_weight.length; i++) {
			System.out.printf("%.2f ",unfac_weight[i]);
		}
		System.out.println();
	}
	
	private static double[] sum_weights(SpliceGraph s, int[] map) {
		Edge[] line_edgelist = s_line.getEdgeList();
		double[] weight_sum = new double[line_edgelist.length];
		Edge[] edge_list = s.getEdgeList();
		Edge[][] matrix = s.getEdges();
		double[] weights = new double[edge_list.length];
		
		// figure out actual weights
		for (int i=0; i<edge_list.length; i++) {
			if (s.is_5prime(edge_list[i])) {
				// initial condition
				weights[i] = edge_list[i].weight;
			}
			else {
				double incoming = 0.0;
				int index = s.getSegIndex(edge_list[i].source);
				for (int j=0; j<matrix.length; j++) {
					if (matrix[j][index] != null) {
						int edge_ind = s.getEdgeIndex(matrix[j][index]);
						incoming += weights[edge_ind];
					}
				}
				weights[i] = incoming*edge_list[i].weight;
			}
		}
		
		for (int i=0; i<map.length; i++) {
			weight_sum[map[i]] += weights[i];
		}
		
		matrix = s_line.getEdges();
		double[] out_totals = new double[matrix.length];
		for (int i=0; i<matrix.length; i++) {
			for (int j=0; j<matrix[i].length; j++) {
				if (matrix[i][j] != null) {
					// i has an outgoing edge to j
					int edge_ind = s_line.getEdgeIndex(matrix[i][j]);
					out_totals[i] += weight_sum[edge_ind];
				}
			}
		}
		
		for (int i=0; i<weight_sum.length; i++) {
			weight_sum[i] /= out_totals[s_line.getSegIndex(line_edgelist[i].source)];
		}
		
		return weight_sum;
	}
	
	private static int[] match_edges(SpliceGraph s, HashMap<Integer,Integer> start_pos,
									 HashMap<Integer,Integer> end_pos) {
		Edge[] edge_list = s.getEdgeList();
		int[] map = new int[edge_list.length];
		
		Edge[][] matrix = s_line.getEdges();
		for (int i=0; i<edge_list.length; i++) {
			int source = end_pos.get(edge_list[i].source.end);   // index of source seg
			int target = start_pos.get(edge_list[i].target.start); // index of target seg
			map[i] = s_line.getEdgeIndex(matrix[source][target]);
		}
		
		return map;
	}
	
	private static void get_values(SpliceGraph s, String filename) {
		Edge[] edge_list = s.getEdgeList();
		try {
			Scanner r = new Scanner(new File(filename));
			String iters = "";
			while (r.hasNextLine()) {
				String line = r.nextLine();
				if (line.startsWith("Converged")) {
					String[] components = line.split("[\\s]+");
					iters = components[2];
					iters += " ";
					break;
				}
			}
			while (r.hasNextLine()) {
				String line = r.nextLine();
				if (line.startsWith(iters)) {
					String[] components = line.split("[\\s]+");
					assert(edge_list.length == (components.length-3));
					for (int i=3; i<components.length; i++) {
						edge_list[i-3].weight = Double.parseDouble(components[i]);
					}
					break;
				}
			}
		}
		catch (IOException ioe) {
			System.out.println("Error reading "+filename);
			System.exit(0);
		}
	}
}