import java.util.*;
import java.io.*;

public class GFF_to_PSG {
    
    private static int get_by_ID(ArrayList<Seg> s, String ID) {
	int retval = -1;
	for (int i=0; i<s.size(); i++) {
	    if (s.get(i).ID.equals(ID)) {
		retval = i;
		break;
	    }
	}
	return retval;
    }
    
    public static void main (String[] args) {
	// at this point I'm not entirely sure where to get sequence data
	// so I'm going to conveniently ignore that for now
	
	if (args.length != 1) {
	    System.out.println("usage: GFF_to_PSG <filename>");
	    System.exit(0);
	}
	
	String filename = args[0];
		
	ArrayList<Seg> segments = new ArrayList<Seg>();
	ArrayList<Edg> edges = new ArrayList<Edg>();
	segments.add(new Seg("start")); // this is our dummy starter segment
	try {
	    ///////////////////////
	    // First: add segments
	    ///////////////////////
	    Scanner r = new Scanner(new File(filename));
	    while (r.hasNextLine()) {
		String[] components = r.nextLine().split("\t");
		// components[3],[4] contain start/end information
		int start = Integer.parseInt(components[3]);
		int end = Integer.parseInt(components[4]);
		int length = Math.abs(end-start);
		String sequence = "";
		for (int i=0; i<length; i++) {
		    sequence += "A";
		}
		// components[8] contains node ID, parent(s) of node
		String[] attributes = components[8].split(";");
		if (!attributes[0].startsWith("ID=")) {
		    System.out.println("Improperly formatted file "+filename);
		    System.exit(0);
		}
		
		// make segment
		if (!components[2].equalsIgnoreCase("graph")) {
		    String ID = attributes[0].substring(3);
		    Seg temp = new Seg(ID);
		    temp.sequence = sequence;
		    segments.add(temp);
		}
	    }
	    
	    ///////////////////////
	    // Second: add edges
	    ///////////////////////
	    r = new Scanner(new File(filename));
	    while (r.hasNextLine()) {
		String[] components = r.nextLine().split("\t");
		String[] attributes = components[8].split(";");
		// components[2] contains parent/child information
		if (components[2].equalsIgnoreCase("graph")) {
		    // this is where we could get sequence info?
		    // not sure where SpliceGrapher is storing its seq data
		}
		else if (components[2].equalsIgnoreCase("parent")) {
		    // connect it to node 0
		    Edg ed = new Edg();
		    ed.source = 0;
		    ed.target = get_by_ID(segments,attributes[0].substring(3));
		    edges.add(ed);
		}
		else if (components[2].equalsIgnoreCase("child")) {
		    // get parent info from components[8]
		    if (attributes.length < 2 || 
			!attributes[1].startsWith("Parent=")) {
			System.out.println("Improperly formatted file "+
					   filename);
			System.exit(0);
		    }
		    String[] parents = attributes[1].substring(7).split(",");
		    if (parents.length > 1) {
			int target = get_by_ID(segments,attributes[0].substring(3));
			for (int i=0; i<parents.length; i++) {
			    Edg ed = new Edg();
			    ed.target = target;
			    ed.source = get_by_ID(segments,parents[i]);
			    edges.add(ed);
			}
		    }
		    else {
			Edg ed = new Edg();
			ed.target = get_by_ID(segments,attributes[0].substring(3));
			ed.source = get_by_ID(segments,parents[0]);
			edges.add(ed);
		    }
		}
	    }
	}
	catch (IOException ioe) {
	    System.out.println("Error reading file "+filename);
	    System.exit(0);
	}
	catch (ArrayIndexOutOfBoundsException oob) {
	    System.out.println("Improperly formatted file "+filename);
	    System.exit(0);
	}
	catch (NumberFormatException nfe) {
	    System.out.println("Improperly formatted file "+filename);
	    System.exit(0);
	}
	System.out.println("File "+filename+"read, closing graph.");
	
	close_graph(segments,edges);
	
	System.out.println("Graph closed, sorting segments.");
	
	///////////////////////
	// Third: Sort segments
	///////////////////////
	ArrayList<Seg> sort_seg = new ArrayList<Seg>();
	ArrayList<Edg> sort_edg = new ArrayList<Edg>();
	sort_seg.add(segments.get(0));
	for(int i=0; i<sort_seg.size(); i++) {
	    int curr_ind = sort_seg.get(i).index;
	    for (int j=0; j<edges.size(); j++) {
		if (edges.get(j).source == curr_ind) {
		    // migrate target
		    if (!sort_seg.contains(segments.get(edges.get(j).target))) {
			sort_seg.add(segments.get(edges.get(j).target));
		    }
		    Edg foo = new Edg();
		    foo.source = i;
		    foo.target = sort_seg.size()-1;
		    foo.weight = edges.get(j).weight;
		    sort_edg.add(foo);
		}
	    }
	}
	
	System.out.println("Segments sorted, writing PSG.");
	
	// We have segments and edges, time to write the file
	write_file(filename,sort_seg,sort_edg);
	System.out.println("PSG complete.");
    }
    
    private static void close_graph(ArrayList<Seg> s, ArrayList<Edg> e) {
	// all of the finishing touches, like adding a dummy end node
	// and figuring out basic weights
	
	// count outdegree
	int[] num_out = new int[s.size()];
	for (int i=0; i<e.size(); i++) {
	    num_out[e.get(i).source]++;
	}
	s.add(new Seg("end"));
	
	for (int i=0; i<num_out.length; i++) {
	    if (num_out[i] > 0) {
		double weight = 1.0/(double)num_out[i];
		for (int j=0; j<e.size(); j++) {
		    if (e.get(j).source == i) {
			e.get(j).weight = weight;
		    }
		}
	    }
	    else {
		// connect it to end dummy node
		Edg ed = new Edg();
		ed.source = i;
		ed.target = s.size()-1;
		ed.weight = 1.0;
		e.add(ed);
	    }
	}
    }
    
    private static void write_file(String filename, ArrayList<Seg> s,
				   ArrayList<Edg> e) {
	// Note: filename will still have .gff suffix
	String newfile = filename.substring(0,filename.length()-4);
	newfile += ".new.psg";
	
	try {
	    BufferedWriter out = new BufferedWriter(new FileWriter(newfile));
	    out.write("[segments]");
	    out.newLine();
	    for (int i=0; i<s.size(); i++) {
		out.write(Integer.toString(i));
		out.write("\t");
		out.write(s.get(i).sequence);
		out.newLine();
	    }
	    out.write("[edges]");
	    out.newLine();
	    for (int i=0; i<e.size(); i++) {
		out.write(Integer.toString(e.get(i).source));
		out.write("\t");
		out.write(Integer.toString(e.get(i).target));
		out.write("\t");
		out.write(String.format("%.5f",e.get(i).weight));
		out.newLine();
	    }
	    out.close();
	}
	catch (IOException ioe) {
	    System.out.println("Error writing to file "+newfile);
	    System.exit(0);
	}
    }
    
}

class Edg {
    // dummy class to help sort these edges out
    
    public int source;
    public int target;
    public double weight;
    
    public Edg() {
	source = -1;
	target = -1;
	weight = 0.0;
    }
}

class Seg {
    // dummy class to help sort these nodes out, since they're not in order
    private static int counter = 0;
    
    public String sequence;
    public String ID;
    public int index;
    
    public Seg(String i) {
	this.sequence = "";
	this.ID = i;
	this.index = counter++;
    }
}