import java.io.*;
import java.util.*;

public class PSGTranslator {

    public static String sequence;
    public static Segment[] segments;
    public static Edge[] edge_list;

    public static void main(String[] args) {
	if (args.length != 1) {
	    System.out.println("usage: PSGTranslator <psg>");
	    System.exit(0);
	}

	String filename = args[0];
	if (!filename.substring(filename.lastIndexOf(".")).equals(".psg")) {
	    System.out.println("Please use a .psg file");
	    System.exit(0);
	}

	String[] lines = getLines(filename);
	int line_num = 0;
	int section_start = -1;
	segments = null;
	edge_list = null;
	sequence = null;
	while (line_num <= lines.length) {
	    if (line_num == lines.length || lines[line_num].startsWith("[")) {
		if (section_start != -1) {
		    String section = lines[section_start].trim();
		    section = section.substring(1,section.length()-1);
		    String[] s_lines = getSectionLines(lines,
						       section_start,line_num);
		    if (section.equalsIgnoreCase("sequence")) {
			sequence = readSequence(s_lines);
		    }
		    else if (section.equalsIgnoreCase("segments")) {
			segments = readSegments(s_lines);
		    }
		    else if (section.equalsIgnoreCase("edges")) {
			edge_list = readEdges(s_lines);
		    }
		}
		section_start = line_num;
	    }
	    line_num++;
	}

	if (segments == null || edge_list == null || sequence == null) {
	    System.out.println("ERROR: Improper PSG format.");
	    System.exit(0);
	}

	filename = filename.substring(0,filename.lastIndexOf("."));
	filename += ".new.psg";

	try {
	    BufferedWriter out = new BufferedWriter(new FileWriter(filename));
	    out.write("[segments]");
	    out.newLine();
	    //String seq = s.getSequence();
	    //Segment[] segs = s.getSegments();
	    for (int i=0; i<segments.length; i++) {
		out.write(i+"\t");
		out.write(sequence.substring(segments[i].start,
					     segments[i].end));
		out.newLine();
	    }
	    out.write("[edges]");
	    out.newLine();
	    //Edge[] edges = s.getEdgeList();
	    for (int i=0; i<edge_list.length; i++) {
		int source = getSegIndex(edge_list[i].source);
		int target = getSegIndex(edge_list[i].target);
		out.write(source+"\t"+target+"\t"+edge_list[i].weight);
		out.newLine();
	    }
	    out.close();
	}
	catch (IOException ioe) {
	    System.out.println("Error writing to file "+filename);
	    System.exit(0);
	}
    }

    private static int getSegIndex(Segment s) {
	for (int i=0; i<segments.length; i++) {
	    if (segments[i] == s) {
		return i;
	    }
	}
	return -1;
    }

    private static String readSequence(String[] a) {
	String retval = "";
	for (int i=0; i<a.length; i++) {
	    retval += a[i];
	}
	return retval;
    }

    private static Segment[] readSegments(String[] a) {
	ArrayList<Segment> retval = new ArrayList<Segment>();
	for (int i=0; i<a.length; i++) {
	    Segment temp = new Segment();
	    String[] s = a[i].split("[\\s]+");
	    temp.start = Integer.parseInt(s[1]);
	    temp.end = Integer.parseInt(s[2]);
	    retval.add(temp);
	}
	return Utility.SegArray(retval);
    }

    private static Edge[] readEdges(String[] a) {
	ArrayList<Edge> retval = new ArrayList<Edge>();
	for (int i=0; i<a.length; i++) {
	    Edge temp = new Edge();
	    String[] s = a[i].split("[\\s]+");
	    temp.weight = Double.parseDouble(s[2]);
	    try {
		temp.source = segments[Integer.parseInt(s[0])];
		temp.target = segments[Integer.parseInt(s[1])];
	    }
	    catch (NullPointerException npe) {
		System.out.println("ERROR: PSG format requires segments"+
				   " to be listed before edges.");
		System.exit(0);
	    }
	    retval.add(temp);
	}
	return Utility.EdgArray(retval);
    }

    private static String[] getLines(String filename) {
	ArrayList<String> lines = new ArrayList<String>();
	try {
	    Scanner r = new Scanner(new File(filename));
	    while (r.hasNextLine()) {
		// discards all comment lines
		String line = r.nextLine();
		line = trimData(line);
		if (line.length() > 0) {
		    lines.add(line);
		}
	    }
	}
	catch (IOException ioe) {
	    System.out.println(filename+" contains an error.");
	    System.exit(0);
	}
	return Utility.StringArray(lines);
    }

    private static String[] getSectionLines(String[] a, int start, 
					    int line_num) {
	ArrayList<String> s_lines = new ArrayList<String>();
	for (int i=start+1; i<line_num; i++) {
	    s_lines.add(a[i]);
	}
	return Utility.StringArray(s_lines);
    }

    // cleans up whitespace, removes comments
    private static String trimData(String s) {
	String r = s.trim();
	if (r.startsWith("#")) {
	    r = "";
	}
	return r;
    }

}